/*  Power Station Industrializer
 *  Copyright (c) 2000 David A. Bartold
 *  Copyright (c) 2003, 2020 Yury Aliaev
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define ALSA_PCM_NEW_HW_PARAMS_API
#define ALSA_PCM_NEW_SW_PARAMS_API

#include <alsa/asoundlib.h>

#include "alsa.h"
#include "interface.h"

const gchar* const default_device = "plughw:0,0";	/* default playback device */
const snd_pcm_format_t format = SND_PCM_FORMAT_S16;	/* sample format */
const unsigned int channels = 1;	/* count of channels */

snd_pcm_t *handle;

static gint device, devices; /* Really tags */
static GtkWidget *cwidget;

static int alsa_open(const gint prate)
{
    unsigned int rrate = prate;
    snd_pcm_uframes_t bufsize;
    int err;

    snd_pcm_hw_params_t *hwparams;
    snd_pcm_sw_params_t *swparams;

    snd_pcm_hw_params_alloca(&hwparams);
    snd_pcm_sw_params_alloca(&swparams);

    if ((err =
	 snd_pcm_open(&handle, conf_get_string(device), SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
	return err;
    }
    if ((err = snd_pcm_hw_params_any(handle, hwparams)) < 0) {
        goto end_err;
    }
    if ((err =
	 snd_pcm_hw_params_set_access(handle, hwparams,
				      SND_PCM_ACCESS_RW_INTERLEAVED)) <
	0) {
        goto end_err;
    }
    if ((err = snd_pcm_hw_params_set_format(handle, hwparams, format)) < 0) {
        goto end_err;
    }
    if ((err = snd_pcm_hw_params_set_channels(handle, hwparams, 1)) < 0) {
        goto end_err;
    }
    if ((err =
	 snd_pcm_hw_params_set_rate_near(handle, hwparams, &rrate,
					 0)) < 0) {
        goto end_err;
    }
    if ((float) rrate / (float) prate > 1.1
	|| (float) rrate / (float) prate < .9) {
	fprintf(stderr, "Supported rate is too far from %i Hz\n", prate);
        err = G_MININT;
        goto end_err;
    }
    if ((err =
	 snd_pcm_hw_params_get_buffer_size_max(hwparams, &bufsize)) < 0) {
        goto end_err;
    }
    if (bufsize < 2 * 2048) {
	fprintf(stderr, "Buffer size is too small, sorry...\n");
        err = G_MININT;
        goto end_err;
    }
    if ((err =
	 snd_pcm_hw_params_set_buffer_size(handle, hwparams, bufsize =
					   2 * 2048)) < 0) {
        goto end_err;
    }
    guint pers = 2;
    if ((err = snd_pcm_hw_params_set_periods_near(handle, hwparams, &pers, NULL)) < 0) {
        goto end_err;
    }
    if ((err = snd_pcm_hw_params(handle, hwparams)) < 0) {
        goto end_err;
    }
    if ((err = snd_pcm_sw_params_current(handle, swparams)) < 0) {
        goto end_err;
    }
    if ((err =
	 snd_pcm_sw_params_set_start_threshold(handle, swparams,
					       bufsize)) < 0) {
        goto end_err;
    }
    if ((err =
	 snd_pcm_sw_params_set_avail_min(handle, swparams, 2048)) < 0) {
        goto end_err;
    }
    if ((err = snd_pcm_sw_params(handle, swparams)) < 0) {
        goto end_err;
    }
    if ((err = snd_pcm_prepare(handle)) < 0)
        goto end_err;

    return 0;
end_err:
    snd_pcm_close(handle);
    return err;
}

static int xrun_recovery(snd_pcm_t * handle, int err)
{
    if (err == -EPIPE) {	/* under-run */
	err = snd_pcm_prepare(handle);
	if (err < 0)
	    fprintf(stderr,
		    "Can't recovery from underrun, prepare failed: %s\n",
		    snd_strerror(err));
	return 0;
    } else if (err == -ESTRPIPE) {
	while ((err = snd_pcm_resume(handle)) == -EAGAIN)
	    sleep(1);		/* wait until the suspend flag is released */
	if (err < 0) {
	    err = snd_pcm_prepare(handle);
	    if (err < 0)
		fprintf(stderr,
			"Can't recovery from suspend, prepare failed: %s\n",
			snd_strerror(err));
	}
	return 0;
    }
    return err;
}

static int alsa_play(gint16 * ptr, int n)
{
    int err = 0;

    while (err <= 0) {
	err = snd_pcm_writei(handle, ptr, n);

	if (err == -EAGAIN)
	    continue;
	if (err < 0) {
	    if (xrun_recovery(handle, err) < 0) {
		return err;
	    }
	    break;		/* skip one period */
	}
    }

    return err;
}

static void alsa_close(void)
{
    snd_pcm_drop(handle);
    snd_pcm_close(handle);
}

static const char *alsa_err(int errnum)
{
    static const char *message =
	N_("PSIndustrializer internal error\nSee stderr for details");

    if (errno == G_MININT)
	return message;

    return snd_strerror(errnum);
}

static GtkWidget*
alsa_get_config_widget(void)
{
    static GtkWidget *box = NULL;

    if(!cwidget) {
        GSList *curdev;
        GtkWidget *thing;

        box = gtk_hbox_new(FALSE, 4);
        thing = gtk_label_new(_("Device:"));
        gtk_box_pack_start(GTK_BOX(box), thing, FALSE, FALSE, 0);
        cwidget = gtk_combo_box_text_new_with_entry();
        for(curdev = conf_get_string_array(devices); curdev; curdev = curdev->next)
            gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(cwidget), curdev->data);
        gui_combo_box_prepend_text_or_set_active(cwidget, conf_get_string(device));
        gtk_box_pack_start(GTK_BOX(box), cwidget, TRUE, TRUE, 0);
        gtk_widget_show_all(box);
    }

    return box;
}

static void
alsa_loadsettings(void)
{
    devices = conf_subscribe_string_array("driver/alsa/", "devices", default_device);
    device = conf_subscribe_string("driver/alsa/", "current_device", default_device);
}

static void
alsa_applysettings(const gboolean apply)
{
    gchar *dev_s = conf_get_string(device);
    /* APPLY means that the config widget is do shown */
    if(apply) {
        g_free(dev_s);
        dev_s = gtk_combo_box_text_get_active_text(GTK_COMBO_BOX_TEXT(cwidget));
        conf_set_string(device, dev_s);
    }
    if(cwidget)
        /* Check if the user adds something new */
        if(!gui_combo_box_prepend_text_or_set_active(cwidget, dev_s)) {
            GSList *dev_list = g_slist_prepend(conf_get_string_array(devices), dev_s);

            conf_set_string_array(devices, dev_list);
        }
}

drv driver_alsa = {
    N_("ALSA"),
    NULL,
    alsa_open,
    alsa_play,
    alsa_close,
    NULL,
    alsa_err,
    alsa_get_config_widget,
    alsa_loadsettings,
    alsa_applysettings
};
