/*  Power Station Industrializer
 *  Copyright (c) 2000 David A. Bartold
 *  Copyrigth (c) 2005-2020 Yury Aliaev
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>

#include <gdk/gdkkeysyms.h>
#ifdef HAVE_OPENGL
#  include <gtk/gtkgl.h>
#  include <GL/gl.h>
#endif

#include "api-wrapper.h"
#include "interface.h"
#include "main.h"
#include "utils.h"

typedef struct {
    GtkWidget *dialog;
    gboolean is_save;
    const gchar *ext;
    gchar path_tag;
    gint (*callback)(PSPhysMod*, const gchar*, gchar**);
} FileDialogObject;

static GtkWidget *AppWindow;

static GSList *active_list = NULL, *inactive_list = NULL;
static GtkWidget *progressbar1;
static GtkWidget *check_decay, *size_label;
static GtkObject *adj_decay;
static GtkWidget *topologynotebook, *height_spinbutton, *circum_spinbutton, *length_spinbutton,
		 *plane_length_spinbutton, *plane_width_spinbutton;
static GtkWidget *act_node_spin[PS_NUM_OBJECTS], *sam_node_spin[PS_NUM_OBJECTS];
static GtkWidget *speed_hscale, *damping_hscale, *actuation_combo, *velocity_hscale, *tension_hscale;
static GtkWidget *prev_cwidget = NULL;
#ifdef HAVE_OPENGL
static GtkWidget *area = NULL;
static float x_angle = 90.0, y_angle = 0.0;
#endif

static PSMetalObj *object = NULL;

static guint	preselected_driver;
static gboolean	autocorrect_ext, overwarning;
static gboolean need_render = TRUE;
static gint conf_autoext, conf_overwrite_warning, conf_rate, conf_exp_rate; /* Tags */
static gint rate, pre_rate, exp_rate, pre_erate;

static PSPhysMod model;


static void
render_object()
{
    ps_metal_obj_free(object);

    switch (model.obj_type) {
    case PS_OBJECT_TUBE:
	object = ps_metal_obj_new_tube(model.height, model.circum, model.tenseness);
	break;

    case PS_OBJECT_ROD:
	object = ps_metal_obj_new_rod(model.length, model.tenseness);
	break;

    case PS_OBJECT_PLANE:
	object =
	    ps_metal_obj_new_plane(model.plane_length, model.plane_width, model.tenseness);
	break;
    default:
        g_assert_not_reached();
    }

#ifdef HAVE_OPENGL
    gtk_widget_queue_draw(area);
#endif
}

#ifdef HAVE_OPENGL
static gint glarea_draw(GtkWidget * widget, GdkEventExpose * event);
static gint glarea_realize(GtkWidget * widget);

static gdouble basex, basey;

static gint glarea_press(GtkWidget * widget, GdkEventMotion * event)
{
    basex = event->x;
    basey = event->y;

    return TRUE;
}

static gint glarea_release(GtkWidget * widget, GdkEventMotion * event)
{
    return TRUE;
}

static void
cube(float x1, float x2, float y1, float y2, float z1, float z2)
{
    /* Face 1 */
    glBegin(GL_TRIANGLE_STRIP);
    glNormal3f(0.0, 0.0, -1.0);
    glVertex3f(x1, y1, z1);
    glVertex3f(x1, y2, z1);
    glVertex3f(x2, y1, z1);
    glVertex3f(x2, y2, z1);
    glEnd();

    /* Face 2 */
    glBegin(GL_TRIANGLE_STRIP);
    glNormal3f(1.0, 0.0, 0.0);
    glVertex3f(x2, y1, z1);
    glVertex3f(x2, y2, z1);
    glVertex3f(x2, y1, z2);
    glVertex3f(x2, y2, z2);
    glEnd();

    /* Face 3 */
    glBegin(GL_TRIANGLE_STRIP);
    glNormal3f(0.0, 0.0, 1.0);
    glVertex3f(x2, y1, z2);
    glVertex3f(x2, y2, z2);
    glVertex3f(x1, y1, z2);
    glVertex3f(x1, y2, z2);
    glEnd();

    /* Face 4 */
    glBegin(GL_TRIANGLE_STRIP);
    glNormal3f(-1.0, 0.0, 0.0);
    glVertex3f(x1, y1, z2);
    glVertex3f(x1, y2, z2);
    glVertex3f(x1, y1, z1);
    glVertex3f(x1, y2, z1);
    glEnd();

    /* Face 5 */
    glBegin(GL_TRIANGLE_STRIP);
    glNormal3f(0.0, -1.0, 0.0);
    glVertex3f(x1, y1, z2);
    glVertex3f(x1, y1, z1);
    glVertex3f(x2, y1, z2);
    glVertex3f(x2, y1, z1);
    glEnd();

    /* Face 6 */
    glBegin(GL_TRIANGLE_STRIP);
    glNormal3f(0.0, 1.0, 0.0);
    glVertex3f(x1, y2, z1);
    glVertex3f(x1, y2, z2);
    glVertex3f(x2, y2, z1);
    glVertex3f(x2, y2, z2);
    glEnd();
}

static void glarea_update(GtkWidget * widget)
{
    int i;
    vector3 *v3;
    float maxx, maxy, maxz;
    float minx, miny, minz;
    float max, min;
    float cenx, ceny, cenz;
    float size;
    float ptsize;

    GLfloat light_ambient[] = { 0.5, 0.5, 0.5, 1.0 };
    GLfloat light_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat light_position[] = { 10.0, 10.0, 10.0, 0.0 };
    GLfloat global_ambient[] = { 0.75, 0.75, 0.75, 1.0 };

    GLfloat red_mat_diffuse[] = { 1.0, 0.0, 0.0, 1.0 };
    GLfloat green_mat_diffuse[] = { 0.0, 1.0, 0.0, 1.0 };
    GLfloat blue_mat_diffuse[] = { 0.0, 0.0, 1.0, 1.0 };
    GLfloat cyan_mat_diffuse[] = { 0.0, 1.0, 1.0, 1.0 };
    GLfloat white_mat_diffuse[] = { 1.0, 1.0, 1.0, 1.0 };

    if (widget == NULL || !gtk_widget_get_realized(widget))
	return;

    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);
    GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
    if (gdk_gl_drawable_gl_begin (gldrawable, glcontext)) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (object != NULL) {
	    minx = maxx = 0.0;
	    miny = maxy = 0.0;
	    minz = maxz = 0.0;

	    for (i = 0; i < object->num_nodes; i++) {
		v3 = &object->nodes[i]->pos;

		if (v3->x < minx)
		    minx = v3->x;
		if (v3->y < miny)
		    miny = v3->y;
		if (v3->z < minz)
		    minz = v3->z;
		if (v3->x > maxx)
		    maxx = v3->x;
		if (v3->y > maxy)
		    maxy = v3->y;
		if (v3->z > maxz)
		    maxz = v3->z;
	    }

	    min = minx;
	    if (miny < min)
		min = miny;
	    if (minz < min)
		min = minz;
	    max = maxx;
	    if (maxy > max)
		max = maxy;
	    if (maxz > max)
		max = maxz;

	    cenx = (minx + maxx) / 2.0;
	    ceny = (miny + maxy) / 2.0;
	    cenz = (minz + maxz) / 2.0;
	    size = (max - min) / 2.0 * 1.2;

	    glEnable(GL_LIGHTING);
	    glEnable(GL_LIGHT0);
	    glDepthRange(0.4, 0.6);

	    glFrontFace(GL_CCW);
	    glDepthFunc(GL_LESS);
	    glEnable(GL_DEPTH_TEST);
	    glCullFace(GL_BACK);
	    glEnable(GL_CULL_FACE);
	    glViewport(0, 0, widget->allocation.width,
		       widget->allocation.height);
	    glMatrixMode(GL_PROJECTION);
	    glLoadIdentity();
	    glFrustum(-1.0, 1.0, -1.0, 1.0, 1.0, 200.0);
	    glScalef(95.0 / size, 95.0 / size, 1.0);

	    glMatrixMode(GL_MODELVIEW);
	    glLoadIdentity();

	    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
	    glLightfv(GL_LIGHT0, GL_POSITION, light_position);

	    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);

	    glTranslatef(0.0, 0.0, -100.0);
	    glRotatef(y_angle, 1.0, 0.0, 0.0);
	    glRotatef(x_angle, 0.0, 1.0, 0.0);
	    glTranslatef(-cenx, -ceny, -cenz);

	    ptsize = model.tenseness * 0.4;
	    if (ptsize > 0.4)
		ptsize = 0.4;

	    for (i = 0; i < object->num_nodes; i++) {
		v3 = &object->nodes[i]->pos;

		if (object->nodes[i]->anchor)
		    glMaterialfv(GL_FRONT, GL_DIFFUSE, red_mat_diffuse);
		else if (i == model.innode) {
		    if (i == model.outnode)
		        glMaterialfv(GL_FRONT, GL_DIFFUSE, cyan_mat_diffuse);
		    else
		        glMaterialfv(GL_FRONT, GL_DIFFUSE, green_mat_diffuse);
		} else if (i == model.outnode)
		    glMaterialfv(GL_FRONT, GL_DIFFUSE, blue_mat_diffuse);
		else
		    glMaterialfv(GL_FRONT, GL_DIFFUSE, white_mat_diffuse);

		cube(v3->x - ptsize, v3->x + ptsize, v3->y - ptsize,
		     v3->y + ptsize, v3->z - ptsize, v3->z + ptsize);
	    }
	}

        if (gdk_gl_drawable_is_double_buffered (gldrawable))
            gdk_gl_drawable_swap_buffers (gldrawable);
        else
            glFlush ();
        gdk_gl_drawable_gl_end (gldrawable);
    }
}

static gint glarea_motion(GtkWidget * widget, GdkEventMotion * event)
{
    /* This code is borrowed from the planetmm GtkGLArea-- sample app. */

    gdouble x, y;
    gfloat xChange, yChange;
    GdkModifierType state;

    /* Calculate the relative motion of the cursor */
    if (event->is_hint) {
	gint tx, ty;
	gdk_window_get_pointer(event->window, &tx, &ty, &state);
	x = (gdouble) tx;
	y = (gdouble) ty;
    } else {
	x = event->x;
	y = event->y;
	state = (GdkModifierType) event->state;
    }

    if (!(state & GDK_BUTTON1_MASK))
	return TRUE;

    xChange = 0.25 * (GLfloat) (basex - x);
    yChange = 0.25 * (GLfloat) (basey - y);
    basex = event->x;
    basey = event->y;

    /* Transform curser motions into object rotations */
    x_angle += xChange;
    y_angle += yChange;

    /* Limit the values of rotation */
    while (x_angle < 0.0)
	x_angle += 360.0;
    while (x_angle > 360.0)
	x_angle -= 360.0;
    while (y_angle < 0.0)
	y_angle += 360.0;
    while (y_angle > 360.0)
	y_angle -= 360.0;

    gtk_widget_queue_draw(widget);

    return TRUE;
}
#endif

static GtkWidget
*opengl_create(gchar * widget_name, gchar * string1,
    gchar * string2, gint int1, gint int2)
{
#ifdef HAVE_OPENGL
    GdkGLContext *glcontext = NULL;
    GdkGLConfig *glconfig = NULL;
    if (glconfig == NULL)
    {
        glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB | GDK_GL_MODE_DEPTH | GDK_GL_MODE_DOUBLE);

        if (glconfig == NULL)
        {
            glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB | GDK_GL_MODE_DEPTH);
        }
    }
    if (glconfig == NULL)
        return gtk_label_new(_("<OpenGL is buggy!>"));
    area = gtk_drawing_area_new ();
    gtk_widget_set_gl_capability (area, glconfig, glcontext, TRUE, GDK_GL_RGBA_TYPE);

    g_signal_connect(GTK_OBJECT(area), "expose_event",
		       G_CALLBACK(glarea_draw), NULL);

    g_signal_connect(GTK_OBJECT(area), "realize",
		       G_CALLBACK(glarea_realize), NULL);

    g_signal_connect(GTK_OBJECT(area), "button_press_event",
		       G_CALLBACK(glarea_press), NULL);

    g_signal_connect(GTK_OBJECT(area), "button_release_event",
		       G_CALLBACK(glarea_release), NULL);

    g_signal_connect(GTK_OBJECT(area), "motion_notify_event",
		       G_CALLBACK(glarea_motion), NULL);

    gtk_widget_set_size_request(area, 300, 300);

    return area;
#else
    return gtk_label_new(_("<OpenGL not supported!>"));
#endif
}

#ifdef HAVE_OPENGL
static gint glarea_draw (GtkWidget * widget, GdkEventExpose * event)
{
    if (event->count > 0)
	return TRUE;

    glarea_update(widget);

    return TRUE;
}

static gint glarea_realize (GtkWidget * widget)
{
    render_object();

    return TRUE;
}
#endif

static void presets_changed()
{
    need_render = TRUE;
}

static gboolean
quit_requested(void)
{
    gtk_main_quit();
    return TRUE;
}

static void
quit_requested_v(void)
{
    gtk_main_quit();
}

static void
update_nodes_range(void)
{
	gint node_min, node_max;

	switch (model.obj_type) {
	case PS_OBJECT_TUBE:
		node_min = model.circum;
		node_max = model.circum * (model.height - 1) - 1;
		break;

	case PS_OBJECT_ROD:
		node_min = 1;
		node_max = model.length - 2;
		break;

	case PS_OBJECT_PLANE:
		node_min = 1;
		node_max = model.plane_length * model.plane_width - 2;
		break;
	default:
		g_assert_not_reached();
	}

	gtk_spin_button_set_range(GTK_SPIN_BUTTON(act_node_spin[model.obj_type]),
		node_min, node_max);
	gtk_spin_button_set_range(GTK_SPIN_BUTTON(sam_node_spin[model.obj_type]),
		node_min, node_max);
}

static void
on_spinbutton_changed(GtkSpinButton *sb, gint *val)
{
	*val = gtk_spin_button_get_value_as_int(sb);
	update_nodes_range();

	/* Reset input and output nodes to their default values */
	switch (model.obj_type) {
	case PS_OBJECT_TUBE:
		model.innode = model.circum + model.circum / 2;
		model.outnode = (model.height - 2) * model.circum;
		break;
	case PS_OBJECT_ROD:
		model.innode = 1;
		model.outnode = model.length - 2;
		break;
	case PS_OBJECT_PLANE:
		model.innode = 1;
		model.outnode = (model.plane_length - 1) * model.plane_width - 1;
		break;
	default:
		g_assert_not_reached();
	}

	gtk_spin_button_set_value(GTK_SPIN_BUTTON(act_node_spin[model.obj_type]), model.innode);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(sam_node_spin[model.obj_type]), model.outnode);
	presets_changed();
	render_object();
}

static void
on_node_spinbutton_changed(GtkSpinButton *sb, gint *val)
{
    const gint newval = gtk_spin_button_get_value_as_int(sb);

    if (model.obj_type == PS_OBJECT_PLANE) {
        /* Corners of the plane are anchors. We should avoid
           setting them as either actuation or sampling point */
        if (newval == model.plane_width - 1 ||
            newval == (model.plane_length - 1 ) * model.plane_width) {
            gint dir = newval > *val ? 1 : -1;
            gtk_spin_button_set_value(sb, newval + dir);
            /* No need to continue, this function
               now will be called once again */
            return;
        }
    }

    *val = newval;
    presets_changed();
	render_object();
}

static void
on_tension_hscale_changed(GtkAdjustment * adj, gpointer user_data)
{
    model.tenseness = gtk_adjustment_get_value(adj);
    presets_changed();
    render_object();
}

static void
on_adj_value_changed(GtkAdjustment * adj, gdouble *val)
{
    *val = gtk_adjustment_get_value(adj);
    presets_changed();
}

static void
on_actuation_comboentry_changed(GtkComboBox * combobox, gpointer user_data)
{
    model.actuation = gtk_combo_box_get_active(combobox);
    presets_changed();
}

static void
on_use_decay_toggled(GtkToggleButton * button, GtkWidget * target)
{
    model.decay_is_used = gtk_toggle_button_get_active(button);
    gtk_widget_set_sensitive(target, model.decay_is_used);
    presets_changed();
}

static void
on_topologynotebook_switch_page (GtkNotebook * notebook,
				gpointer * page,
				gint page_num, gpointer user_data)
{
    model.obj_type = page_num;
    model.innode =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(act_node_spin[page_num]));
    model.outnode =
        gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(sam_node_spin[page_num]));
    render_object();
    presets_changed();
}

static void
set_status_message(gchar * str)
{
    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progressbar1), str);
}

static void
set_percent(gfloat percent, gpointer data)
{
    gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(progressbar1), percent);
    while(gtk_events_pending())
        gtk_main_iteration();
}

static void
gui_set_values()
{
    gtk_range_set_value(GTK_RANGE(tension_hscale), model.tenseness);
    gtk_range_set_value(GTK_RANGE(speed_hscale), model.speed);
    gtk_range_set_value(GTK_RANGE(damping_hscale), model.damping);
    gtk_range_set_value(GTK_RANGE(velocity_hscale), model.velocity);

    gtk_combo_box_set_active(GTK_COMBO_BOX(actuation_combo), model.actuation);
    switch (model.obj_type) {
    case PS_OBJECT_TUBE:
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(height_spinbutton), (gfloat)model.height);
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(circum_spinbutton), (gfloat)model.circum);
        break;
    case PS_OBJECT_ROD:
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(length_spinbutton), (gfloat)model.length);
        break;
    case PS_OBJECT_PLANE:
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(plane_length_spinbutton), (gfloat)model.plane_length);
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(plane_width_spinbutton), (gfloat)model.plane_width);
        break;
    default:
        g_assert_not_reached();
    }
}

static void
gui_set_size_label (gfloat value)
{
    static gchar *size_str;

    size_str = g_strdup_printf(_("Real length: %.1f s"), value);
    gtk_label_set_text(GTK_LABEL(size_label), size_str);
    g_free(size_str);
}

static void
driver_selected(GtkComboBox * combobox, GtkBox *c_area)
{
    GtkWidget *cwidget;
    drv *p_driver;

    preselected_driver = gtk_combo_box_get_active(combobox);
    p_driver = g_slist_nth_data(driver_list, preselected_driver);

    if (p_driver && p_driver->get_config_widget)
        cwidget = p_driver->get_config_widget();
    else
        cwidget = NULL;
    if (cwidget != prev_cwidget) {
        if (prev_cwidget) {
            g_object_ref(prev_cwidget);
            gtk_container_remove(GTK_CONTAINER(c_area), prev_cwidget);
        }
        if (cwidget) {
            gtk_box_pack_start(c_area, cwidget, FALSE, FALSE, 4);
            gtk_box_reorder_child(c_area, cwidget, 2);
        }
        prev_cwidget = cwidget;
    }
}

static void descr_fill(drv * ldriver, GtkWidget *combo)
{
    gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(combo), ldriver->description);
}

static void checkbutton_changed(GtkToggleButton * tbutton, gboolean *value)
{
    *value = gtk_toggle_button_get_active(tbutton);
}

static void
rate_changed(GtkComboBox *w, GtkTreeModel *mdl, gint* newrate)
{
    GtkTreeIter iter;

    if (!gtk_combo_box_get_active_iter(w, &iter))
        return;
    gtk_tree_model_get(mdl, &iter, 0, newrate, -1);
}

static void
playrate_changed(GtkComboBox *w, GtkTreeModel *mdl)
{
    rate_changed(w, mdl, &pre_rate);
}

static void
exprate_changed(GtkComboBox *w, GtkTreeModel *mdl)
{
    rate_changed(w, mdl, &pre_erate);
}

typedef struct {
    guint value;
    gint number;
} compare_data;

static gboolean
compare_func(GtkTreeModel* model, GtkTreePath* path, GtkTreeIter* iter, gpointer data)
{
    compare_data* cmp_data = (compare_data*)data;
    guint cur_val;

    gtk_tree_model_get(model, iter, 0, &cur_val, -1);

    if (cur_val == cmp_data->value) {
        gint* indices = gtk_tree_path_get_indices(path);

        cmp_data->number = indices[0];
        return TRUE; /* The desired element is found */
    }

    return FALSE;
}

static gboolean
set_active_combo_item(GtkWidget* combobox, GtkTreeModel* model, guint item)
{
    compare_data cmp_data;

    cmp_data.value = item;
    cmp_data.number = -1;
    gtk_tree_model_foreach(model, compare_func, &cmp_data);

    if (cmp_data.number >= 0) {
        gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), cmp_data.number);
        return TRUE;
    }

    return FALSE;
}

static GtkWidget*
create_rate_combo(const gchar* const title,
    GtkListStore* *ls,
    GtkWidget *tobox,
    const gchar* const tooltip)
{
    GtkWidget *thing, *box;
    GtkCellRenderer *cell;
    const gint freqs[] = {44100, 48000, 96000};
    gint i;

    box = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(tobox), box, FALSE, FALSE, 1);
    thing = gtk_label_new(_(title));
    gtk_box_pack_start(GTK_BOX(box), thing, FALSE, FALSE, 0);

    *ls = gtk_list_store_new(1, G_TYPE_UINT);
    thing = gtk_combo_box_new_with_model(GTK_TREE_MODEL(*ls));
    for (i = 0; i < sizeof(freqs) / sizeof(freqs[0]); i++) {
        GtkTreeIter iter;

        gtk_list_store_append(*ls, &iter);
        gtk_list_store_set(*ls, &iter, 0, freqs[i], -1);
    }
    g_object_unref(*ls);
    cell = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(thing), cell, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(thing), cell, "text", 0, NULL);
    gtk_box_pack_end(GTK_BOX(box), thing, TRUE, TRUE, 0);
    if (tooltip)
        gtk_widget_set_tooltip_text(box, _(tooltip));

    return thing;
}

static void setup_dialog(void)
{
    static GtkWidget *combo, *check_ext, *check_overwrite, *c_area, *rate_combo, *exp_rate_combo;
    static GtkWidget *setup_window = NULL;
    static GtkListStore *rate_ls, *exp_rate_ls;
    gint response;
    guint i;

    if (!setup_window) {
        GtkWidget *thing, *box;

        setup_window = gtk_dialog_new_with_buttons(_("Industrializer configuration"),
            GTK_WINDOW(AppWindow), GTK_DIALOG_MODAL,
            GTK_STOCK_OK, GTK_RESPONSE_OK,
            GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);
        gtk_window_set_resizable(GTK_WINDOW(setup_window), FALSE);
        c_area = gtk_dialog_get_content_area(GTK_DIALOG(setup_window));

        box = gtk_hbox_new(FALSE, 4);
        gtk_box_pack_start(GTK_BOX(c_area), box, FALSE, FALSE, 1);
        thing = gtk_label_new(_("Sound output:"));
        gtk_box_pack_start(GTK_BOX(box), thing, FALSE, FALSE, 0);
        combo = gtk_combo_box_text_new();
        g_slist_foreach(driver_list, (GFunc) descr_fill, combo);
        gtk_box_pack_end(GTK_BOX(box), combo, TRUE, TRUE, 0);
        g_signal_connect(combo, "changed", G_CALLBACK(driver_selected), c_area);

        if (driver && driver->get_config_widget) {
            prev_cwidget = driver->get_config_widget();
            gtk_box_pack_start(GTK_BOX(c_area), prev_cwidget, FALSE, FALSE, 4);
        }

        rate_combo = create_rate_combo(N_("Playback rate:"), &rate_ls, c_area, NULL);
        g_signal_connect(rate_combo, "changed", G_CALLBACK(playrate_changed), rate_ls);

        thing = gtk_hseparator_new();
        gtk_box_pack_start(GTK_BOX(c_area), thing, FALSE, FALSE, 2);

        check_ext = gtk_check_button_new_with_label(_("Add extension to files without it"));
        gtk_box_pack_start(GTK_BOX(c_area), check_ext, FALSE, FALSE, 0);
        g_signal_connect(check_ext, "toggled",  G_CALLBACK(checkbutton_changed),
            &autocorrect_ext);

        check_overwrite = gtk_check_button_new_with_label(_("File overwrite warning"));
        gtk_box_pack_start(GTK_BOX(c_area), check_overwrite, FALSE, FALSE, 0);
        g_signal_connect(check_overwrite, "toggled", G_CALLBACK(checkbutton_changed),
            &overwarning);

        exp_rate_combo = create_rate_combo(N_("Export rate:"), &exp_rate_ls, c_area,
            N_("Discretization rate for export to WAV file"));
        g_signal_connect(exp_rate_combo, "changed", G_CALLBACK(exprate_changed), exp_rate_ls);
        gtk_widget_show_all(c_area);
    }

    gtk_combo_box_set_active(GTK_COMBO_BOX(combo), psi_get_current_driver());
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check_ext),
        autocorrect_ext = conf_get_bool(conf_autoext)); /* innocent C hack ;-) */
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check_overwrite),
        overwarning = conf_get_bool(conf_overwrite_warning));
    set_active_combo_item(rate_combo, GTK_TREE_MODEL(rate_ls), rate);
    set_active_combo_item(exp_rate_combo, GTK_TREE_MODEL(exp_rate_ls), exp_rate);

    response = gtk_dialog_run(GTK_DIALOG(setup_window));
    gtk_widget_hide(setup_window);
    if (response == GTK_RESPONSE_OK) {
        conf_set_bool(conf_autoext, autocorrect_ext);
        conf_set_bool(conf_overwrite_warning, overwarning);
        rate = pre_rate;
        conf_set_int(conf_rate, rate);
        psi_set_driver(preselected_driver, rate);
        exp_rate = pre_erate;
        conf_set_int(conf_exp_rate, exp_rate);
    }
    /* We may change other drivers' settings before selecting the desired driver.
       Let's reset these settings to the last remembered values */
    for(i = 0; i < g_slist_length(driver_list); i++) {
        drv *cur_driver = g_slist_nth_data(driver_list, i);
        if(cur_driver->applysettings)
            /* For the selected driver we apply settings if the response is ok */
            cur_driver->applysettings(i == preselected_driver && response == GTK_RESPONSE_OK);
    }
}

static void
create_AboutWindow (void)
{
    const gchar* const authors[] = {
        "David A. Bartold <foxx@mail.utexas.edu>",
        "Yury Aliaev <mutab0r@rambler.ru>",
        "Conrad Parker <conradp@cse.unsw.edu.au>",
        "Wladimir van der Laan <laanwj@gmail.com>",
        NULL};
    GdkPixbuf *pbuf = gdk_pixbuf_new_from_file(PREFIX"/share/"PACKAGE"/power_station_logo.xpm", NULL);
    gchar *license = g_strdup_printf(_("%s is free software: you can redistribute it and/or modify it under the terms "
        "of the GNU General Public License as published by the Free Software Foundation; either version 2 "
        "of the License, or (at your option) any later version.\n\n"
        "%s is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the "
        "implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General "
        "Public License for more details.\n\n"
        "You should have received a copy of the GNU General Public License along with %s"
        ". If not, see: http://www.gnu.org/licenses/"), PACKAGE_NAME, PACKAGE_NAME, PACKAGE_NAME);

    gtk_show_about_dialog(GTK_WINDOW(AppWindow), "program_name", PACKAGE_NAME,
        "version", PACKAGE_VERSION, "authors", authors, "license", license,
        "wrap-license", TRUE, "copyright",
        _("Copyright \302\251 2000-2005 David A. Bartold\n"
        "Copyright \302\251 2014 Wladimir J. van der Laan\n"
        "Copyright \302\251 2020-2021 Yury G. Aliaev\n\n"),
        "comments", _("Power Station Industrializer\nA Percussion Synthesizer\n\n"),
        "logo", pbuf, NULL);
    g_free(license);
    g_object_unref(pbuf);
}

static void set_active(const gboolean active)
{
    GSList *l;

    for (l = active_list; l; l = l->next)
        gtk_widget_set_sensitive(GTK_WIDGET(l->data), active);
    for (l = inactive_list; l; l = l->next)
        gtk_widget_set_sensitive(GTK_WIDGET(l->data), !active);
}

static void
start_render(gint rrate)
{
    static gint lastrate = 0;
    gsize size;

    if (!need_render && rrate == lastrate)
        return;

    set_status_message(_("Rendering..."));
    set_active(FALSE);

    size = do_render(&model, rrate);
    if (size) {
        need_render = FALSE;
        lastrate = rrate;
    }

    set_active(TRUE);
    set_status_message(_("Ready..."));
    gui_set_size_label((gfloat) size / rrate);
    set_percent(0.0, NULL);
}

static void
play_sound()
{
    GSList *l;

    start_render(rate);
    for (l = inactive_list; l; l = l->next)
        gtk_widget_set_sensitive(GTK_WIDGET(l->data), TRUE);
    if (trigger_play(&model, rate))
        /* Playing is finished, not retrigged */
        for (l = inactive_list; l; l = l->next)
            gtk_widget_set_sensitive(GTK_WIDGET(l->data), FALSE);
}

static gint
trigger_save(PSPhysMod *m, const gchar *fname, gchar **errmsg)
{
    start_render(exp_rate);
    return save_wav(m, fname, errmsg, exp_rate);
}

static gint
load_pres_wrapper(PSPhysMod *m, const gchar *fname, gchar **errmsg)
{
    gint retval = load_pres(m, fname, errmsg);

    gtk_notebook_set_current_page(GTK_NOTEBOOK(topologynotebook), m->obj_type);
    gui_set_values();
    presets_changed();

    return retval;
}

static gboolean
gui_ok_cancel_dialog (GtkWidget **dialog, GtkWidget *parent, const gchar *title, const gchar *msg)
{
    gint response;

    if (!*dialog) {
        *dialog = gtk_message_dialog_new(GTK_WINDOW(parent), GTK_DIALOG_MODAL,
            GTK_MESSAGE_QUESTION, GTK_BUTTONS_OK_CANCEL, NULL);
        gtk_window_set_title(GTK_WINDOW(*dialog), _(title));
    }
    gtk_message_dialog_set_markup(GTK_MESSAGE_DIALOG(*dialog), _(msg));
    response = gtk_dialog_run(GTK_DIALOG(*dialog));
    gtk_widget_hide(*dialog);

    return (response == GTK_RESPONSE_OK);
}

static void
file_operation_clicked (GtkWidget *widget, FileDialogObject *fdo)
{
    gint response;
    gboolean fop_do;

    /* Repeat until either user selects an appropriate file name or rejects the operation */
    do {
        fop_do = TRUE;
        response = gtk_dialog_run(GTK_DIALOG(fdo->dialog));

        if(response == GTK_RESPONSE_OK) {
            gchar *fname = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fdo->dialog));

            if(fdo->is_save) {
                static GtkWidget *overwrite_dialog = NULL;

                if(conf_get_bool(conf_autoext)) {
                    gchar *name_buf = g_path_get_basename(fname);

                    /* File can begin with '.'. One-letter name does need an extension */
                    if(strlen(name_buf) == 1 || !strchr(&name_buf[1], '.')) {
                        gchar *fname_new = g_strconcat(fname, ".", fdo->ext, NULL);

                        g_free(fname);
                        fname = fname_new;
                    }
                    g_free(name_buf);
                }
                if(conf_get_bool(conf_overwrite_warning) && g_file_test(fname, G_FILE_TEST_EXISTS))
                    fop_do = gui_ok_cancel_dialog(&overwrite_dialog, fdo->dialog,
                        N_("File overwriting query"),
                        N_("Warning! File exists!\nWould you like to overwrite it?"));
            }
            if(fop_do) {
                gchar *old_path = conf_get_string(fdo->path_tag);
                gchar *path, *path1, *errbuf;
                gint retval;

                path = g_path_get_dirname(fname);
                if(old_path)
                    g_free(old_path);
                path1 = g_filename_to_utf8(path, -1, NULL, NULL, NULL);
                conf_set_string(fdo->path_tag, g_strconcat(path1, "/", NULL));
                g_free(path1);
                g_free(path);

                retval = fdo->callback(&model, fname, &errbuf);
                if(retval) {
                    static GtkWidget *error_dia = NULL;

                    retval != -1 ? gui_errno_dialog(errbuf, retval)
                                 : gui_error_dialog(&error_dia, errbuf);
                    g_free(errbuf);
                }
            }

            g_free(fname);
        }
    } while(!fop_do);
    gtk_widget_hide(fdo->dialog);
}

static FileDialogObject*
gui_create_FileSelector(
    const gchar *title,
    gint path_tag,
    const gchar *ext,
    const gchar *pattern, 
    const gchar *filter_name,
    const gboolean is_save,
    gint (*callback)(PSPhysMod*, const gchar*, gchar**))
{
    gchar *glib_path, *path;
    GtkFileFilter *filter = gtk_file_filter_new();

    FileDialogObject* fdo = g_new(FileDialogObject, 1);

    fdo->dialog = gtk_file_chooser_dialog_new(_(title), GTK_WINDOW(AppWindow),
        is_save ? GTK_FILE_CHOOSER_ACTION_SAVE : GTK_FILE_CHOOSER_ACTION_OPEN,
        GTK_STOCK_OK, GTK_RESPONSE_OK,
        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);
    gtk_container_set_border_width(GTK_CONTAINER(fdo->dialog), 10);
    gtk_window_set_modal(GTK_WINDOW(fdo->dialog), TRUE);

    gtk_file_filter_set_name(filter, _(filter_name));
    gtk_file_filter_add_pattern(filter, pattern);
    gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(fdo->dialog), filter);
    gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(fdo->dialog), filter);

    filter = gtk_file_filter_new();
    gtk_file_filter_set_name(filter, _("All files"));
    gtk_file_filter_add_pattern(filter, "*");
    gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(fdo->dialog), filter);

    path = conf_get_string(path_tag);
    if(path) {
	glib_path = g_filename_from_utf8(path, -1, NULL, NULL, NULL);
	gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(fdo->dialog), glib_path);
	g_free(glib_path);
    }

    fdo->ext = ext;
    fdo->is_save = is_save;
    fdo->callback = callback;
    fdo->path_tag = path_tag;

    return fdo;
}

static void
psi_model_init(PSPhysMod *m)
{
    m->percent_callback = set_percent;

    m->obj_type = PS_OBJECT_TUBE;
    m->decay_is_used = FALSE;
    m->decay_value = -50.0;

    m->height = 10;
    m->circum = 5;
    m->length = 5;
    m->plane_length = 7;
    m->plane_width = 9;
    m->innode = 7;
    m->outnode = 40;
    m->tenseness = 4.0;
    m->speed = 0.2;
    m->damping = 0.05;
    m->actuation = 0;
    m->velocity = 1.0;
}

static GtkWidget*
create_control(const gchar *title,
	const gdouble lower,
	const gdouble upper,
	const gdouble step_increment,
	const gdouble page_increment,
	GtkWidget *table,
	const gint order,
	GCallback func,
	gdouble *parameter)
{
	GtkObject *adj = gtk_adjustment_new(*parameter, lower, upper,
		step_increment, page_increment, 0.0);
	GtkWidget *widget = gtk_label_new(_(title));

	gtk_widget_show(widget);
	gtk_table_attach(GTK_TABLE(table), widget, 0, 1, order, order + 1, 0, 0, 0, 0);

	widget = gtk_spin_button_new(GTK_ADJUSTMENT(adj), step_increment, 3);
	gtk_widget_show(widget);
	gtk_table_attach(GTK_TABLE(table), widget, 2, 3, order, order + 1,
		GTK_FILL, GTK_FILL, 0, 0);

	widget = gtk_hscale_new(GTK_ADJUSTMENT(adj));
	gtk_widget_show(widget);
	g_signal_connect(adj, "value-changed", func, parameter);
	gtk_table_attach(GTK_TABLE(table), widget, 1, 2, order, order + 1,
		GTK_FILL, GTK_FILL, 0, 0);
	gtk_scale_set_digits(GTK_SCALE(widget), 3);

	return widget;
}

static void stop_clicked(void)
{
    model.stop = TRUE;
}

void gui_create_AppWindow(guint current_driver)
{
    typedef enum {
        BUTTON_ALWAYS,
        BUTTON_ACTIVE,
        BUTTON_INACTIVE
    } StockButtonType;

    struct StockButton
    {
        const gchar *id;
        const gchar *tooltip;
        void (*cb)();
        gpointer data;
        guint key;
        GdkModifierType mods;
        StockButtonType type;
    };

    static FileDialogObject *save_dial, *save_pres_dial, *load_pres_dial;
    GtkWidget *vbox3;
    GtkWidget *hbox3;
    GtkWidget *vbox6;
    GtkWidget *table;
    GtkWidget *length_hscale;
    GtkWidget *openglframe;
    GtkWidget *opengl;
    GtkWidget *frame1;
    GtkWidget *hbox4, *thing, *toolbar;
    GtkObject *adj;
    GtkToolItem *titem;
    gchar *size_str;
    GtkAccelGroup *accel_group;
    gint i, pres_path, sample_path;

    psi_model_init(&model);

    conf_autoext = conf_subscribe_bool(config->old_mode ? "behaviour/" : "settings/",
        "auto_ext", TRUE);
    if(config->old_mode)
        conf_change_path(conf_autoext, "settings/");
    conf_overwrite_warning = conf_subscribe_bool(config->old_mode ? "behaviour/" : "settings/",
        "overwrite_warning", TRUE);
    if(config->old_mode)
        conf_change_path(conf_overwrite_warning, "settings/");
    conf_rate = conf_subscribe_int("settings/", "rate", 48000);
    rate = conf_get_int(conf_rate);
    conf_exp_rate = conf_subscribe_int("settings/", "export_rate", 48000);
    exp_rate = conf_get_int(conf_exp_rate);

    AppWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(AppWindow),
			 _("Power Station Industrializer"));
    gtk_window_set_resizable(GTK_WINDOW(AppWindow), FALSE);
    gtk_window_set_icon_from_file(GTK_WINDOW(AppWindow), PREFIX"/share/pixmaps/"PACKAGE".png", NULL);

    accel_group = gtk_accel_group_new();
    gtk_window_add_accel_group(GTK_WINDOW(AppWindow), accel_group);

    vbox3 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox3);
    gtk_container_add(GTK_CONTAINER(AppWindow), vbox3);

    toolbar = gtk_toolbar_new();
    gtk_box_pack_start(GTK_BOX(vbox3), toolbar, FALSE, FALSE, 0);

    sample_path = conf_subscribe_string("paths/", "sample_path", NULL);
    save_dial = gui_create_FileSelector(N_("Save .WAV file"),
        sample_path, "wav", "*.[wW][aA][vV]",
        N_("Microsoft RIFF (*.wav)"), TRUE, trigger_save);
    if(config->old_mode) {
        /* Programming is like sex: one mistake and you have to support 
           it for life. This is for compatibility with older versions 
           of PSI. */
        pres_path = conf_subscribe_string("paths/", "instr_path", NULL);
        conf_rename(pres_path, "presets_path");
    } else
        pres_path = conf_subscribe_string("paths/", "presets_path", NULL);
    save_pres_dial = gui_create_FileSelector(N_("Save presets to file"),
        pres_path, "psip", "*.[pP][sS][iI][pP]",
        N_("PSIndustrializer presets (*.psip)"), TRUE, save_pres);
    load_pres_dial = gui_create_FileSelector(N_("Load presets from file"),
        pres_path, "psip", "*.[pP][sS][iI][iIpP]",
        N_("PSIndustrializer presets (*.psii, *.psip)"), FALSE, load_pres_wrapper);

    struct StockButton sb[] =
    {{GTK_STOCK_OPEN, N_("Open Presets (Ctrl + O)"), file_operation_clicked, load_pres_dial, 'o', GDK_CONTROL_MASK, BUTTON_ACTIVE},
     {GTK_STOCK_SAVE, N_("Save Presets (Ctrl + S)"), file_operation_clicked, save_pres_dial, 's', GDK_CONTROL_MASK, BUTTON_ALWAYS},
     {GTK_STOCK_MEDIA_PLAY, N_("Play Rendered Sound (space)"), play_sound, NULL, GDK_KEY_space, 0, BUTTON_ACTIVE},
     {GTK_STOCK_MEDIA_STOP, N_("Stop Rendering / Playback (Ctrl + B)"), stop_clicked, NULL, 'b', GDK_CONTROL_MASK, BUTTON_INACTIVE},
     {GTK_STOCK_ABOUT, N_("About PSIndustializer (Ctrl + H)"), create_AboutWindow, NULL, 'h', GDK_CONTROL_MASK, BUTTON_ALWAYS},
     {GTK_STOCK_PROPERTIES, N_("Program Setup (Ctrl + U)"), setup_dialog, NULL, 'u', GDK_CONTROL_MASK, BUTTON_ALWAYS},
     {GTK_STOCK_QUIT, N_("Close Program (Ctrl + Q)"), quit_requested_v, NULL, 'q', GDK_CONTROL_MASK, BUTTON_ALWAYS},
     {NULL}
    };

    for(i = 0; sb[i].id; i++) {
        titem = gtk_tool_button_new_from_stock(sb[i].id);
        gtk_tool_item_set_tooltip_text(titem, _(sb[i].tooltip));
        gtk_toolbar_insert(GTK_TOOLBAR(toolbar), titem, i);
        g_signal_connect(GTK_WIDGET(titem), "clicked",
            G_CALLBACK(sb[i].cb), sb[i].data);
        gtk_widget_add_accelerator(GTK_WIDGET(titem), "clicked", accel_group,
            sb[i].key, sb[i].mods, 0);
        if (sb[i].type == BUTTON_ACTIVE)
            active_list = g_slist_prepend(active_list, titem);
        else if (sb[i].type == BUTTON_INACTIVE)
            inactive_list = g_slist_prepend(inactive_list, titem);
    }

    const int sizes[] = {22, 32};
    for(i = 0; i < sizeof(sizes) / sizeof(sizes[0]); i++) {
        gchar *icon_path = g_strdup_printf(PREFIX"/share/"PACKAGE"/save-wav-%i.png", sizes[i]);
        GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file(icon_path, NULL);

        gtk_icon_theme_add_builtin_icon("save_wav", gdk_pixbuf_get_width(pixbuf), pixbuf);
        g_object_unref(G_OBJECT(pixbuf));
        g_free(icon_path);
    }

    titem = gtk_tool_button_new(NULL, _("Save WAV"));
    gtk_tool_button_set_icon_name(GTK_TOOL_BUTTON(titem), "save_wav");
    gtk_tool_item_set_tooltip_text(titem, _("Save rendered sound as WAV (Ctrl + W)"));
    gtk_widget_add_accelerator(GTK_WIDGET(titem), "clicked", accel_group, 'w', GDK_CONTROL_MASK, 0);
    g_signal_connect(GTK_WIDGET(titem), "clicked",
        G_CALLBACK(file_operation_clicked), save_dial);
    gtk_toolbar_insert(GTK_TOOLBAR(toolbar), titem, 4);
    active_list = g_slist_prepend(active_list, titem);

    gtk_widget_show_all(toolbar);
    set_active(TRUE);

    hbox3 = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox3);
    gtk_box_pack_start(GTK_BOX(vbox3), hbox3, TRUE, TRUE, 0);

    vbox6 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox6);
    gtk_box_pack_start(GTK_BOX(hbox3), vbox6, TRUE, TRUE, 0);

    topologynotebook = gtk_notebook_new();
    gtk_widget_show(topologynotebook);
    gtk_box_pack_start(GTK_BOX(vbox6), topologynotebook, TRUE, TRUE, 0);

    table = gtk_table_new(4, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 10);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    thing = gtk_label_new(_("Tube Circumference"));
    gtk_table_attach(GTK_TABLE(table), thing, 0, 1, 1, 2, 0, 0, 0, 0);

    height_spinbutton =
	gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(model.height, 3, 30, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), height_spinbutton, 1, 2, 0, 1, 0, 0, 0, 0);
    g_signal_connect(height_spinbutton, "value-changed",
		     G_CALLBACK(on_spinbutton_changed), &model.height);

    thing = gtk_label_new(_("Tube Height"));
    gtk_table_attach(GTK_TABLE(table), thing, 0, 1, 0, 1, 0, 0, 0, 0);

    circum_spinbutton =
	gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(model.circum, 3, 30, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), circum_spinbutton, 1, 2, 1, 2, 0, 0, 0, 0);
    g_signal_connect(circum_spinbutton, "value-changed",
		     G_CALLBACK(on_spinbutton_changed), &model.circum);

    thing = gtk_label_new(_("Actuation Node"));
    gtk_table_attach(GTK_TABLE(table), thing, 2, 3, 0, 1, 0, 0, 0, 0);
	thing = act_node_spin[PS_OBJECT_TUBE] =
		gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(model.circum + model.circum / 2,
			model.circum, model.circum * (model.height - 1) - 1, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), thing, 3, 4, 0, 1, 0, 0, 0, 0);
    g_signal_connect(thing, "value-changed",
		     G_CALLBACK(on_node_spinbutton_changed), &model.innode);

    thing = gtk_label_new(_("Sampling Node"));
    gtk_table_attach(GTK_TABLE(table), thing, 2, 3, 1, 2, 0, 0, 0, 0);
	thing = sam_node_spin[PS_OBJECT_TUBE] =
		gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new((model.height - 2) * model.circum,
			model.circum, model.circum * (model.height - 1) - 1, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), thing, 3, 4, 1, 2, 0, 0, 0, 0);
    g_signal_connect(thing, "value-changed",
		     G_CALLBACK(on_node_spinbutton_changed), &model.outnode);

    gtk_widget_show_all(table);
    gtk_notebook_append_page(GTK_NOTEBOOK(topologynotebook), table,
        gtk_label_new(_("Tube")));

    table = gtk_table_new(4, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 10);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    thing = gtk_label_new(_("Rod Length"));
    gtk_table_attach(GTK_TABLE(table), thing, 0, 1, 0, 1, 0, 0, 0, 0);

    length_spinbutton =
	gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(model.length, 3, 200, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), length_spinbutton, 1, 2, 0, 1, 0, 0, 0, 0);
    g_signal_connect(length_spinbutton, "value-changed",
		     G_CALLBACK(on_spinbutton_changed), &model.length);

    thing = gtk_label_new(_("Actuation Node"));
    gtk_table_attach(GTK_TABLE(table), thing, 2, 3, 0, 1, 0, 0, 0, 0);
	thing = act_node_spin[PS_OBJECT_ROD] =
		gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(1,
			1, model.length - 2, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), thing, 3, 4, 0, 1, 0, 0, 0, 0);
    g_signal_connect(thing, "value-changed",
		     G_CALLBACK(on_node_spinbutton_changed), &model.innode);

    thing = gtk_label_new(_("Sampling Node"));
    gtk_table_attach(GTK_TABLE(table), thing, 2, 3, 1, 2, 0, 0, 0, 0);
	thing = sam_node_spin[PS_OBJECT_ROD] =
		gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(model.length - 2,
			1, model.length - 2, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), thing, 3, 4, 1, 2, 0, 0, 0, 0);
    g_signal_connect(thing, "value-changed",
		     G_CALLBACK(on_node_spinbutton_changed), &model.outnode);

    gtk_widget_show_all(table);
    gtk_notebook_append_page(GTK_NOTEBOOK(topologynotebook), table,
        gtk_label_new(_("Rod")));

    table = gtk_table_new(4, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), 10);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    thing = gtk_label_new(_("Plane Length"));
    gtk_table_attach(GTK_TABLE(table), thing, 0, 1, 0, 1, 0, 0, 0, 0);

    thing = gtk_label_new(_("Plane Width"));
    gtk_table_attach(GTK_TABLE(table), thing, 0, 1, 1, 2, 0, 0, 0, 0);

    plane_length_spinbutton =
	gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(model.plane_length,
		3, 30, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), plane_length_spinbutton, 1, 2, 0, 1, 0, 0, 0, 0);
    g_signal_connect(plane_length_spinbutton, "value-changed",
		     G_CALLBACK(on_spinbutton_changed), &model.plane_length);

    plane_width_spinbutton =
	gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(model.plane_width,
		3, 39, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), plane_width_spinbutton, 1, 2, 1, 2, 0, 0, 0, 0);
    g_signal_connect(plane_width_spinbutton, "value-changed",
		     G_CALLBACK(on_spinbutton_changed), &model.plane_width);

    thing = gtk_label_new(_("Actuation Node"));
    gtk_table_attach(GTK_TABLE(table), thing, 2, 3, 0, 1, 0, 0, 0, 0);
	thing = act_node_spin[PS_OBJECT_PLANE] =
		gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(1,
			1, model.plane_length * model.plane_width - 2, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), thing, 3, 4, 0, 1, 0, 0, 0, 0);
    g_signal_connect(thing, "value-changed",
		     G_CALLBACK(on_node_spinbutton_changed), &model.innode);

    thing = gtk_label_new(_("Sampling Node"));
    gtk_table_attach(GTK_TABLE(table), thing, 2, 3, 1, 2, 0, 0, 0, 0);
	thing = sam_node_spin[PS_OBJECT_PLANE] =
		gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new((model.plane_length - 1) * model.plane_width - 1,
			1, model.plane_length * model.plane_width - 2, 1, 10, 0)), 1, 0);
    gtk_table_attach(GTK_TABLE(table), thing, 3, 4, 1, 2, 0, 0, 0, 0);
    g_signal_connect(thing, "value-changed",
		     G_CALLBACK(on_node_spinbutton_changed), &model.outnode);

    gtk_widget_show_all(table);
    gtk_notebook_append_page(GTK_NOTEBOOK(topologynotebook), table,
        gtk_label_new(_("Plane")));

    table = gtk_table_new(7, 3, FALSE);
    gtk_widget_show(table);
    gtk_box_pack_start(GTK_BOX(vbox6), table, TRUE, TRUE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(table), 10);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

	thing = gtk_label_new(_("Type of Actuation"));
	gtk_widget_show(thing);
	gtk_table_attach(GTK_TABLE(table), thing, 0, 1, 3, 4, 0, 0, 0, 0);

	thing = gtk_label_new(_("Sample Length"));
	gtk_widget_show(thing);
	gtk_table_attach(GTK_TABLE(table), thing, 0, 1, 5, 6, 0, 0, 0, 0);

    thing = gtk_label_new(_("Final Decay"));
    gtk_widget_show(thing);
    gtk_table_attach(GTK_TABLE(table), thing, 0, 1, 6, 7,
		     (GtkAttachOptions) (0), (GtkAttachOptions) (0), 0, 0);

	tension_hscale = create_control(N_("Spring Tension"),
		0.1, 16.0, 0.005, 0.1, table, 0,
		G_CALLBACK(on_tension_hscale_changed), &model.tenseness);
	gtk_widget_set_can_focus(tension_hscale, TRUE);
	gtk_widget_set_can_default(tension_hscale, FALSE);

	speed_hscale = create_control(N_("Speed"),
		0, 0.5, 0.01, 0.02, table, 1,
		G_CALLBACK(on_adj_value_changed), &model.speed);
	damping_hscale = create_control(N_("Damping time"),
		0.005, 0.5, 0.001, 0.01, table, 2,
		G_CALLBACK(on_adj_value_changed), &model.damping);

    actuation_combo = gtk_combo_box_text_new();
    gtk_widget_show(actuation_combo);
    gtk_table_attach(GTK_TABLE(table), actuation_combo, 1, 2, 3, 4,
		     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		     (GtkAttachOptions) (0), 0, 0);
    gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(actuation_combo), _("Compression"));
    gtk_combo_box_text_append_text(GTK_COMBO_BOX_TEXT(actuation_combo), _("Perpendicular Hit"));
    gtk_combo_box_set_active(GTK_COMBO_BOX(actuation_combo), model.actuation);
    g_signal_connect(actuation_combo, "changed",
		     G_CALLBACK(on_actuation_comboentry_changed), NULL);

	velocity_hscale = create_control(N_("Velocity"),
		0.0, 1.0, 0.1, 0.05,  table, 4,
		G_CALLBACK(on_adj_value_changed), &model.velocity);

    length_hscale =
	gtk_hscale_new(GTK_ADJUSTMENT
		       (adj = gtk_adjustment_new(1, 0, 10, 0.1, 0.2, 0)));
    gtk_widget_show(length_hscale);
    g_signal_connect(adj, "value-changed",
		     G_CALLBACK(on_adj_value_changed), &sample_length);
    gtk_table_attach(GTK_TABLE(table), length_hscale, 1, 2, 5, 6,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (GTK_FILL), 0, 0);
    gtk_scale_set_digits(GTK_SCALE(length_hscale), 3);

    hbox4 = gtk_hbox_new(FALSE, 0);
    check_decay = gtk_check_button_new();
    gtk_widget_set_tooltip_text(check_decay,
		 _("Check this if you want to stop rendering at given decay"));
    gtk_widget_show(check_decay);
    gtk_box_pack_start(GTK_BOX(hbox4), check_decay, FALSE, TRUE, 0);

    thing =
	gtk_hscale_new(GTK_ADJUSTMENT
		       (adj_decay =
			gtk_adjustment_new(-50.0, -80.0, -20.0, 1.0, 5.0, 0)));
    gtk_widget_set_tooltip_text(thing,
		 _("Relative signal amplitude at which the rendering is stopped"));
    gtk_widget_set_sensitive(thing, 0);
    gtk_widget_show(thing);
    gtk_box_pack_start(GTK_BOX(hbox4), thing, TRUE, TRUE, 0);
    gtk_scale_set_digits(GTK_SCALE(thing), 2);
    g_signal_connect(adj_decay, "value-changed",
		     G_CALLBACK(on_adj_value_changed), &model.decay_value);
    g_signal_connect(check_decay, "toggled",
		     G_CALLBACK(on_use_decay_toggled), thing);
    gtk_widget_show(hbox4);
    gtk_table_attach(GTK_TABLE(table), hbox4, 1, 2, 6, 7,
		     (GtkAttachOptions) (GTK_FILL),
		     (GtkAttachOptions) (GTK_FILL), 0, 0);

    openglframe = gtk_aspect_frame_new(_("3D View"), 0.5, 0.5, 1, FALSE);
    gtk_widget_show(openglframe);
    gtk_box_pack_start(GTK_BOX(hbox3), openglframe, TRUE, TRUE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(openglframe), 4);

    opengl = opengl_create("opengl", NULL, NULL, 0, 0);
    gtk_widget_show(opengl);
    gtk_container_add(GTK_CONTAINER(openglframe), opengl);
#ifdef HAVE_OPENGL
    gtk_widget_set_can_focus(opengl, TRUE);
    gtk_widget_set_can_default(opengl, FALSE);
    gtk_widget_set_events(opengl,
			  GDK_EXPOSURE_MASK | GDK_POINTER_MOTION_HINT_MASK
			  | GDK_BUTTON1_MOTION_MASK | GDK_BUTTON_PRESS_MASK
			  | GDK_BUTTON_RELEASE_MASK);
#endif

    frame1 = gtk_frame_new(_("Status"));
    gtk_widget_show(frame1);
    gtk_box_pack_start(GTK_BOX(vbox3), frame1, TRUE, TRUE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(frame1), 10);

    hbox4 = gtk_hbox_new(FALSE, 10);
    gtk_widget_show(hbox4);
    gtk_container_add(GTK_CONTAINER(frame1), hbox4);
    gtk_container_set_border_width(GTK_CONTAINER(hbox4), 10);

    progressbar1 = gtk_progress_bar_new();
    gtk_widget_show(progressbar1);
    gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progressbar1), _("Ready..."));
    gtk_box_pack_start(GTK_BOX(hbox4), progressbar1, TRUE, TRUE, 0);

    size_str = g_strdup_printf(_("Real length: %.1f s"), 0.0);
    size_label = gtk_label_new(size_str);
    gtk_widget_show(size_label);
    g_free(size_str);
    gtk_box_pack_start(GTK_BOX(hbox4), size_label, FALSE, FALSE, 0);

    g_signal_connect(AppWindow, "delete-event",
		     G_CALLBACK(quit_requested), NULL);
    g_signal_connect(topologynotebook, "switch-page",
		     G_CALLBACK(on_topologynotebook_switch_page), NULL);
    g_signal_connect(actuation_combo, "changed",
		     G_CALLBACK(on_actuation_comboentry_changed), NULL);

    psi_set_driver(current_driver, rate);
    gtk_widget_show(AppWindow);
}

typedef struct _str_cmp_data {
    const gchar *str;
    gboolean success;
    GtkWidget *combobox;
} str_cmp_data;

static gboolean
str_cmp_func(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
    gchar *item_str = NULL;
    str_cmp_data *scd = (str_cmp_data*)data;

    gtk_tree_model_get(model, iter, 0, &item_str, -1);
    if (!item_str)
        return TRUE; /* Aborting due to error */

    if (!g_ascii_strcasecmp(item_str, scd->str)) {
        scd->success = TRUE;
        gtk_combo_box_set_active_iter(GTK_COMBO_BOX(scd->combobox), iter);
        g_free(item_str);
        return TRUE;
    }

    g_free(item_str);
    return FALSE;
}

gboolean gui_combo_box_prepend_text_or_set_active(GtkWidget *combobox, const gchar *text)
{
    str_cmp_data scd;

    GtkTreeModel *model = gtk_combo_box_get_model(GTK_COMBO_BOX(combobox));

    scd.str = text;
    scd.success = FALSE;
    scd.combobox = combobox;
    gtk_tree_model_foreach(model, str_cmp_func, &scd);

    if (!scd.success) {
        gtk_combo_box_text_prepend_text(GTK_COMBO_BOX_TEXT(combobox), text);
        gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0);
    }

    return scd.success;
}

void
gui_message_dialog(GtkWidget** dialog,
    const gchar* text,
    GtkMessageType type,
    const gchar* title)
{
    if (!*dialog) {
        *dialog = gtk_message_dialog_new(GTK_WINDOW(AppWindow), GTK_DIALOG_MODAL, type,
            GTK_BUTTONS_CLOSE, "%s", text);
        gtk_window_set_title(GTK_WINDOW(*dialog), _(title));
    } else
        gtk_message_dialog_set_markup(GTK_MESSAGE_DIALOG(*dialog), text);

    gtk_dialog_run(GTK_DIALOG(*dialog));
    gtk_widget_hide(*dialog);
}

void
gui_errno_dialog(const gchar *text, const gint err)
{
    gchar *buf = g_strdup_printf("%s: %s", text, g_strerror(err));
    static GtkWidget *dialog = NULL;

    if (!dialog) {
        dialog = gtk_message_dialog_new(GTK_WINDOW(AppWindow), GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR,
                                        GTK_BUTTONS_CLOSE, "%s", buf);
        gtk_window_set_title(GTK_WINDOW(dialog), _("Error!"));
    } else
        gtk_message_dialog_set_markup(GTK_MESSAGE_DIALOG(dialog), buf);

    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_hide(dialog);
    g_free(buf);
}
