/*  Power Station Industrializer
 *  Copyright (c) 2000 David A. Bartold
 *  Copyright (c) 2020 Yury G. Aliaev
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef WIN32
#  include <io.h>
#  include <windows.h>
#  include <windowsx.h>
#  include <mmsystem.h>
#else
#  include <math.h>
#endif

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <audiofile.h>

#include "api-wrapper.h"
#include "utils.h"
#include "xml-parser.h"

static gsize size;
gdouble sample_length = 1.0;

static const gchar pres_ver[] = "0.1.0";

gint16 *samples = NULL;

static const gchar *types[] = {"tube", "rod", "plane"};
#define OBJ_NUM (sizeof(types) / sizeof(types[0]))

static int double_to_s16(double d)
{
    int out;

    if (d >= 1.0)
	out = 32767;
    else if (d <= -1.0)
	out = -32768;
    else
	out = (int) ((d + 1.0) * 32768.0 - 32768.0);

    return out;
}

gsize
do_render(PSPhysMod *model, gint rrate)
{
    int i;
    gfloat decay;

    static double *data;//!!! Get rid of samples, convert data->s16 on the fly
    static unsigned int alloc_length = 0;

    size = (int) (rrate * sample_length);
    if (size > alloc_length) {
	data = g_renew(double, data, size);
	samples = g_renew(gint16, samples, size);
	alloc_length = size;
    }

    decay = model->decay_is_used ? model->decay_value : 0.0;
	size = ps_metal_obj_render(model, rrate, size, data, decay, NULL);

    for (i = 0; i < size; i++)
	samples[i] = double_to_s16(data[i]);

    return size;
}

gboolean
trigger_play(PSPhysMod *model, const gint prate)
{
    if (samples != NULL && driver != NULL) {
#ifdef WIN32
        HWAVEOUT out;
        WAVEFORMATEX format;
        LPWAVEHDR hdr;

        format.wFormatTag = WAVE_FORMAT_PCM;
        format.nChannels = 1;
        format.nSamplesPerSec = prate;
        format.nBlockAlign = 2;
        format.nAvgBytesPerSec =
            format.nSamplesPerSec * format.nBlockAlign;
        format.wBitsPerSample = 16;
        format.cbSize = 0;

        waveOutOpen(&out, WAVE_MAPPER, &format, 0, 0, CALLBACK_NULL);

        hdr = (LPWAVEHDR) GlobalAllocPtr(GMEM_MOVEABLE | GMEM_SHARE,
                                         (DWORD) sizeof(WAVEHDR));

        hdr->lpData =
            GlobalAllocPtr(GMEM_MOVEABLE | GMEM_SHARE, size * 2);
        hdr->dwBufferLength = size * 2;
        hdr->dwFlags = 0;
        memcpy(hdr->lpData, samples, size * 2);
        waveOutPrepareHeader(out, hdr, sizeof(WAVEHDR));
        waveOutWrite(out, hdr, sizeof(WAVEHDR));
        while (!(hdr->dwFlags & WHDR_DONE)) {
            Sleep(100);
        }
        waveOutUnprepareHeader(out, hdr, sizeof(WAVEHDR));
        GlobalFreePtr(hdr->lpData);
        GlobalFree(hdr);

        waveOutClose(out);

#else
        static gboolean is_playing = FALSE;
        gint n;
        static gint nbytes;
        static gint16 *ptr;

        ptr = samples;
        nbytes = size;
        if (is_playing)
            /* Playing is already retrigged by the above lines */
            return FALSE;

        is_playing = TRUE;
        model->stop = FALSE;

        if (driver->open) {
            gint err;

            if ((err = driver->open(prate)) < 0) {
                psi_driver_errmessage(err);
                return TRUE;
            }
        }

        while (nbytes > 0 && !model->stop) {
            if (nbytes > 2048)
                n = 2048;
            else
                n = nbytes;
            n = driver->play(ptr, n);
            if (n < 0) {
                psi_driver_errmessage(n);
                break;
            }
            if (n > 0) {
                ptr += n;
                nbytes -= n;
            }
            while(gtk_events_pending())
                gtk_main_iteration();
        }
        if (driver->close)
            driver->close();
        is_playing = FALSE;
#endif

    }
    return TRUE;
}

gint
save_wav(PSPhysMod *model, const gchar *fname, gchar **errmsg, const gint erate)
{
    gint retcode = 0;

    if (samples != NULL) {
#ifdef WIN32
	/* Microsoft is such a b*stard... they couldn't have made this
	   operation more opaque. */

	HMMIO wav;
	MMIOINFO info;
	MMCKINFO out, outRiff;
	PCMWAVEFORMAT format;
	gint32 total = size * 2;
	gint32 i;

	format.wf.wFormatTag = WAVE_FORMAT_PCM;
	format.wf.nChannels = 1;
	format.wf.nSamplesPerSec = erate;
	format.wf.nBlockAlign = 2;
	format.wf.nAvgBytesPerSec =
	    format.wf.nSamplesPerSec * format.wf.nBlockAlign;
	format.wBitsPerSample = 16;

	unlink(fname);
	wav =
	    mmioOpen(fname, NULL,
		     MMIO_ALLOCBUF | MMIO_READWRITE | MMIO_CREATE);

	outRiff.fccType = mmioFOURCC('W', 'A', 'V', 'E');
	outRiff.cksize = 0;
	mmioCreateChunk(wav, &outRiff, MMIO_CREATERIFF);

	out.ckid = mmioFOURCC('f', 'm', 't', ' ');
	out.cksize = sizeof(PCMWAVEFORMAT);
	mmioCreateChunk(wav, &out, 0);
	mmioWrite(wav, (HPSTR) & format, sizeof(PCMWAVEFORMAT));

	mmioAscend(wav, &out, 0);

	out.ckid = mmioFOURCC('d', 'a', 't', 'a');
	out.cksize = 0;
	mmioCreateChunk(wav, &out, 0);
	mmioGetInfo(wav, &info, 0);

	for (i = 0; i < total; i++) {
	    if (info.pchNext == info.pchEndWrite) {
		info.dwFlags |= MMIO_DIRTY;
		mmioAdvance(wav, &info, MMIO_WRITE);
	    }

	    *((BYTE *) info.pchNext)++ = ((BYTE *) samples)[i];
	}

	info.dwFlags |= MMIO_DIRTY;
	mmioSetInfo(wav, &info, 0);
	mmioAscend(wav, &out, 0);
	mmioAscend(wav, &outRiff, 0);
	mmioClose(wav, 0);
#else
	AFfilehandle wav;
	AFfilesetup setup;

	setup = afNewFileSetup();
	afInitFileFormat(setup, AF_FILE_WAVE);
	afInitSampleFormat(setup, AF_DEFAULT_TRACK, AF_SAMPFMT_TWOSCOMP,
			   16);
	afInitByteOrder(setup, AF_DEFAULT_TRACK,
			AF_BYTEORDER_LITTLEENDIAN);
	afInitChannels(setup, AF_DEFAULT_TRACK, 1);
	afInitRate(setup, AF_DEFAULT_TRACK, erate);

    errno = 0;
	wav = afOpenFile(fname, "w", setup);
	if (wav != NULL) {
	    afWriteFrames(wav, AF_DEFAULT_TRACK, samples, size);
	    afCloseFile(wav);
	} else {
	    retcode = errno;
	    *errmsg = g_strdup_printf(_("Could not write file %s"), fname);
	}

	afFreeFileSetup(setup);
	}
#endif
    return retcode;
}

gint
save_pres(PSPhysMod *model, const gchar* fname, gchar **errmsg)
{
    xmlpContext	*pres = xmlp_new_doc(fname, "psip", pres_ver);

    xmlp_set_string(pres, "", "type", types[model->obj_type]);

    switch(model->obj_type) {
    case PS_OBJECT_TUBE:
	xmlp_set_int(pres, "", "height", model->height);
	xmlp_set_int(pres, "", "circumference", model->circum);
	break;
    case PS_OBJECT_ROD:
	xmlp_set_int(pres, "", "length", model->length);
	break;
    case PS_OBJECT_PLANE:
	xmlp_set_int(pres, "", "length", model->plane_length);
	xmlp_set_int(pres, "", "width", model->plane_width);
	break;
    default:
	break;
    }
    
    xmlp_set_double(pres, "", "tension", model->tenseness);
    xmlp_set_double(pres, "", "speed", model->speed);
    xmlp_set_double(pres, "", "damping", model->damping);
    xmlp_set_int(pres, "", "actuation", model->actuation);
    xmlp_set_double(pres, "", "velocity", model->velocity);
    
    xmlp_sync(pres);//!!! Error handling?
    xmlp_free(pres);

    return 0;
}

gint
load_pres(PSPhysMod *model, const gchar *fname, gchar **errmsg)
{
    guint i, itype = 0;
    gboolean success = TRUE;
    gchar *type;
    xmlpContext *pres = xmlp_get_doc(fname, "psip", pres_ver);

    if(!pres) /* Maybe an old file */
        pres = xmlp_get_doc(fname, "psiinstr", pres_ver);

    if(!pres) {
        *errmsg = g_strdup_printf(_("Could not open file %s, not a PSI presets or parsing error"), fname);
        return -1;
    }

    type = xmlp_get_string_default(pres, "", "type", "tube");
    for(i = 0; i < OBJ_NUM; i++)
        if(!g_ascii_strcasecmp(type, types[i])) {
            itype = i;
            break;
        }
    g_free(type);
    model->obj_type = itype;

    switch(itype) {
    case PS_OBJECT_TUBE:
        success &= xmlp_get_int(pres, "", "height", &model->height);
        success &= xmlp_get_int(pres, "", "circumference", &model->circum);
        break;
    case PS_OBJECT_ROD:
        success &= xmlp_get_int(pres, "", "length", &model->length);
        break;
    case PS_OBJECT_PLANE:
        success &= xmlp_get_int(pres, "", "length", &model->plane_length);
        success &= xmlp_get_int(pres, "", "width", &model->plane_width);
        break;
    default:
        g_assert_not_reached();
    }

    success &= xmlp_get_double(pres, "", "tension", &model->tenseness);
    success &= xmlp_get_double(pres, "", "speed", &model->speed);
    success &= xmlp_get_double(pres, "", "damping", &model->damping);
    success &= xmlp_get_int(pres, "", "actuation", &model->actuation);
    success &= xmlp_get_double(pres, "", "velocity", &model->velocity);
    xmlp_free(pres);

    if(!success) {
        *errmsg = g_strdup(_("Incorrect or incomplete presets file. Some values may be missing."));
        return -1;
    } else
        return 0;
}
